// This file was generated by gir (https://github.com/gtk-rs/gir)
// from gir-files (https://github.com/gtk-rs/gir-files)
// DO NOT EDIT

use crate::ffi;
use glib::{prelude::*, translate::*};

glib::wrapper! {
    #[doc(alias = "PangoCairoFontMap")]
    pub struct FontMap(Interface<ffi::PangoCairoFontMap>) @requires pango::FontMap;

    match fn {
        type_ => || ffi::pango_cairo_font_map_get_type(),
    }
}

impl FontMap {
    pub const NONE: Option<&'static FontMap> = None;

    #[doc(alias = "pango_cairo_font_map_get_default")]
    #[doc(alias = "get_default")]
    #[allow(clippy::should_implement_trait)]
    pub fn default() -> pango::FontMap {
        unsafe { from_glib_none(ffi::pango_cairo_font_map_get_default()) }
    }
}

pub trait PangoCairoFontMapExt: IsA<FontMap> + 'static {
    #[doc(alias = "pango_cairo_font_map_get_resolution")]
    #[doc(alias = "get_resolution")]
    fn resolution(&self) -> f64 {
        unsafe { ffi::pango_cairo_font_map_get_resolution(self.as_ref().to_glib_none().0) }
    }

    #[doc(alias = "pango_cairo_font_map_set_resolution")]
    fn set_resolution(&self, dpi: f64) {
        unsafe {
            ffi::pango_cairo_font_map_set_resolution(self.as_ref().to_glib_none().0, dpi);
        }
    }
}

impl<O: IsA<FontMap>> PangoCairoFontMapExt for O {}
